      subroutine gam2(np,gam,hm,tm,hv,tv,nlag,nx,ny,ndir,ixd,iyd,
     +           vr,tmin,tmax,nvarg,ivtail,ivhead,ivtype,nvar,nvario)
C%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
C                                                                      %
C Copyright (C) 1992 Stanford Center for Reservoir Forecasting.  All   %
C rights reserved.  Distributed with: C.V. Deutsch and A.G. Journel.   %
C ``GSLIB: Geostatistical Software Library and User's Guide,'' Oxford  %
C University Press, New York, 1992.                                    %
C                                                                      %
C The programs in GSLIB are distributed in the hope that they will be  %
C useful, but WITHOUT ANY WARRANTY.  No author or distributor accepts  %
C responsibility to anyone for the consequences of using them or for   %
C whether they serve any particular purpose or work at all, unless he  %
C says so in writing.  Everyone is granted permission to copy, modify  %
C and redistribute the programs in GSLIB, but only under the condition %
C that this notice and the above copyright notice remain intact.       %
C                                                                      %
C%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
c-----------------------------------------------------------------------
c
c                Variogram of Data on a 2-D Regular Grid
c                ***************************************
c
c This subroutine computes any of eight different measures of spatial
c continuity for regular spaced 2-D data.  Missing values are allowed
c and the grid need not be square.
c
c
c
c INPUT VARIABLES:
c
c   nlag             Number of lags to calculate
c   nx               Number of units in x
c   ny               Number of units in y
c   ndir             Number of directions to consider
c   ixd(ndir)        X indicator of direction - number of X grid columns
c                      that must be shifted to move from a node on the
c                      grid to the next nearest node on the grid which
c                      lies on the directional vector.
c   iyd(ndir)        Y indicator of direction - similar to ixd, number
c                      of grid lines that must be shifted to nearest
c                      node which lies on the directional vector
c   nv               The number of variables
c   vr(nx,ny,nv)     Three dimensional array of data
c   tmin,tmax        Trimming limits
c   nvarg            Number of variograms to compute
c   ivtail(nvarg)    Variable for the tail of each variogram
c   ivhead(nvarg)    Variable for the head of each variogram
c   ivtype(nvarg)    Type of variogram to compute:
c                      1. semivariogram
c                      2. cross-semivariogram
c                      3. covariance
c                      4. correlogram
c                      5. general relative semivariogram
c                      6. pairwise relative semivariogram
c                      7. semivariogram of logarithms
c                      8. rodogram
c                      9. madogram
c                     10. indicator semivariogram: an indicator variable
c                         is constructed in the main program.
c
c
c
c OUTPUT VARIABLES:  The following arrays are ordered by direction,
c                    then variogram, and finally lag, i.e.,
c                      iloc = (id-1)*nvarg*nlag+(iv-1)*nlag+il
c
c   np()             Number of pairs
c   gam()            semivariogram, covariance, correlogram,... value
c   hm()             Mean of the tail data
c   tm()             Mean of the head data
c   hv()             Variance of the tail data
c   tv()             Variance of the head data
c
c
c
c Original:  A.G. Journel                                           1978
c Revisions: B.E. Buxton                                       Apr. 1982
c            C.V. Deutsch                                      July 1990
c-----------------------------------------------------------------------
      include   'gam2.inc'
      integer*4 nlag,nx,ny,ndir,nvarg,nvar,i,nvario
      real*8    vr(nx,ny,nvar)
      real*8    gam(nvario),hm(nvario),tm(nvario),hv(nvario),
     +          tv(nvario),np(nvario)
      real*8    tmin,tmax
      real*8    ixd(ndir),iyd(ndir),ivtail(nvarg),ivhead(nvarg),
     +          ivtype(nvarg)
 
c
c Initialize the summation arrays for each direction, variogram, and lag
c
      nsiz = ndir*nvarg*nlag
      if(nsiz.gt.MXDLV) then
            write(*,*) 'ERROR: available storage in gam2 = ',MXDLV
            write(*,*) '       requested storage         = ',nsiz
            stop
      endif
      do 1 i=1,nsiz
            np(i)  = 0.
            gam(i) = 0.0
            hm(i)  = 0.0
            tm(i)  = 0.0
            hv(i)  = 0.0
            tv(i)  = 0.0
 1    continue
c
c First fix the location of a seed point on the grid (ix,iy):
c
      do 2 ix=1,nx
      do 2 iy=1,ny
c
c For the fixed seed point, loop through all directions:
c
            do 3 id=1,ndir
              ixinc = ixd(id)
              iyinc = iyd(id)
              ix1   = ix
              iy1   = iy
c
c For this direction, loop over all the lags:
c
              do 4 il=1,nlag
c
c Check to be sure that the point being considered is still in the
c grid - if not, then finished with this direction:
c
                ix1 = ix1 + ixinc
                if(ix1.lt.1.or.ix1.gt.nx) go to 3
                iy1 = iy1 + iyinc
                if(iy1.lt.1.or.iy1.gt.ny) go to 3
c
c For this direction and lag, loop over all variograms:
c
                do 5 iv=1,nvarg
                  it = ivtype(iv)
c
c Get the head value, skip this value if missing:
c
                  i     = ivhead(iv)
                  vrt   = vr(ix,iy,i)
                  if(vrt.lt.tmin.or.vrt.ge.tmax) go to 5
c
c Get the tail value, skip this value if missing:
c
                  i     = ivtail(iv)
                  vrh   = vr(ix1,iy1,i)
                  if(vrh.lt.tmin.or.vrh.ge.tmax) go to 5
c
c Need increment for the cross semivariogram only:
c
                  if(it.eq.2) then
                        i     = ivhead(iv)
                        vrtpr = vr(ix1,iy1,i)
                        if(vrtpr.lt.tmin.or.vrtpr.ge.tmax) go to 5
                        i     = ivtail(iv)
                        vrhpr = vr(ix,iy,i)
                        if(vrhpr.lt.tmin.or.vrhpr.ge.tmax) go to 5
                  endif
c
c We have an acceptable pair, therefore accumulate all the statistics
c that are required for the variogram:
c
                  i      = (id-1)*nvarg*nlag+(iv-1)*nlag+il
                  np(i)  = np(i) + 1.
                  tm(i)  = tm(i) + dble(vrt)
                  hm(i)  = hm(i) + dble(vrh)
c
c Choose the correct variogram type and keep relevant sums:
c
                  if(it.le.1.or.it.ge.10) then
                    gam(i) = gam(i) + dble((vrh-vrt)*(vrh-vrt))
                  else if(it.eq.2) then
                    gam(i) = gam(i) + dble((vrhpr-vrh)*(vrt-vrtpr))
                  else if(it.eq.3) then
                    gam(i) = gam(i) +  dble(vrh*vrt)
                  else if(it.eq.4) then
                    gam(i) = gam(i) +  dble(vrh*vrt)
                    hv(i)  = hv(i)  +  dble(vrh*vrh)
                    tv(i)  = tv(i)  +  dble(vrt*vrt)
                  else if(it.eq.5) then
                    gam(i) = gam(i) + dble((vrh-vrt)*(vrh-vrt))
                  else if(it.eq.6) then
                    if((vrt+vrh).lt.EPSLON) then
                        np(i)  = np(i) - 1.
                        tm(i)  = tm(i) - vrt
                        hm(i)  = hm(i) - vrh
                    else
                        gamma  = 2.0*(vrt-vrh)/(vrt+vrh)
                        gam(i) = gam(i) + dble(gamma*gamma)
                    endif
                  else if(it.eq.7) then
                    if(vrt.lt.EPSLON.or.vrh.lt.EPSLON) then
                        np(i)  = np(i) - 1.
                        tm(i)  = tm(i) - dble(vrt)
                        hm(i)  = hm(i) - dble(vrh)
                    else
                        gamma  = alog(vrt)-alog(vrh)
                        gam(i) = gam(i) + dble(gamma*gamma)
                    endif
                  else if(it.eq.8) then
                    gam(i) = gam(i) + dble(sqrt(abs(vrt-vrh)))
                  else if(it.eq.9) then
                    gam(i) = gam(i) + dble(abs(vrt-vrh))
                  endif
 5              continue
 4            continue
 3          continue
 2    continue
c
c Get average values for gam, hm, tm, hv, and tv, then compute
c the correct "variogram" measure:
c
      do 6 id=1,ndir
      do 6 iv=1,nvarg
      do 6 il=1,nlag
            i = (id-1)*nvarg*nlag+(iv-1)*nlag+il
            if(np(i).eq.0.) go to 6
            rnum   = np(i)
            gam(i) = gam(i) / dble(rnum)
            hm(i)  = hm(i)  / dble(rnum)
            tm(i)  = tm(i)  / dble(rnum)
            hv(i)  = hv(i)  / dble(rnum)
            tv(i)  = tv(i)  / dble(rnum)
c
c  1. report the semivariogram rather than variogram
c  2. report the cross-semivariogram rather than variogram
c  3. the covariance requires "centering"
c  4. the correlogram requires centering and normalizing
c  5. general relative requires division by lag mean
c  6. report the semi(pairwise relative variogram)
c  7. report the semi(log variogram)
c  8. report the semi(rodogram)
c  9. report the semi(madogram)
c
            it = ivtype(iv)
            if(it.le.2) then
                  gam(i) = 0.5 * gam(i)
            else if(it.eq.3) then
                  gam(i) = gam(i) - hm(i)*tm(i)
            else if(it.eq.4) then
                  hv(i)  = sqrt(hv(i)-hm(i)*hm(i))
                  tv(i)  = sqrt(tv(i)-tm(i)*tm(i))
                  if((hv(i)*tv(i)).lt.EPSLON) then
                        gam(i) = 0.0
                  else
                        gam(i) =(gam(i)-hm(i)*tm(i))/(hv(i)*tv(i))
                  endif
c
c Square "hv" and "tv" so that we return the variance:
c
                  hv(i)  = hv(i)*hv(i)
                  tv(i)  = tv(i)*tv(i)
            else if(it.eq.5) then
                  htave  = 0.5*real(hm(i)+tm(i))
                  htave  = htave   *   htave
                  if(htave.lt.EPSLON) then
                        gam(i) = 0.0
                  else
                        gam(i) = gam(i)/dble(htave)
                  endif
            else if(it.ge.6) then
                  gam(i) = 0.5 * gam(i)
            endif
 6    continue
      return
      end
