      subroutine gamv2u(np,dis,gam,hm,tm,hv,tv,nd,x,y,u,v,tmin,
     +             tmax,nlag,xlag,xltol,ndir,azm,atol,bandw,
     +		   nvarg,ivtail,ivhead,ivtype,nvar,nvario,option)
C%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
C                                                                      %
C Copyright (C) 1992 Stanford Center for Reservoir Forecasting.  All   %
C rights reserved.  Distributed with: C.V. Deutsch and A.G. Journel.   %
C ``GSLIB: Geostatistical Software Library and User's Guide,'' Oxford  %
C University Press, New York, 1992.                                    %
C                                                                      %
C The programs in GSLIB are distributed in the hope that they will be  %
C useful, but WITHOUT ANY WARRANTY.  No author or distributor accepts  %
C responsibility to anyone for the consequences of using them or for   %
C whether they serve any particular purpose or work at all, unless he  %
C says so in writing.  Everyone is granted permission to copy, modify  %
C and redistribute the programs in GSLIB, but only under the condition %
C that this notice and the above copyright notice remain intact.       %
C                                                                      %
C%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
c-----------------------------------------------------------------------
c
c              Variogram of 2-D Irregularly Spaced Data
c              ****************************************
c
c This subroutine computes a variety of spatial continuity measures of a
c set of irregularly spaced data.  The user can specify any combination
c of direct and cross variograms using any of eight "variogram" measures
c
c
c
c INPUT VARIABLES:
c
c   nd               Number of data (no missing values)
c   x(nd)            X coordinates of the data
c   y(nd)            Y coordinates of the data
c   nv               The number of variables
c   vr(nd,nv)        Data values
c   tmin,tmax        Trimming limits
c   nlag             Number of lags to calculate
c   xlag             Length of the unit lag
c   xltol            Distance tolerance (if <0 then set to xlag/2)
c   ndir             Number of directions to consider
c   azm(ndir)        Azimuth angle of direction (measured positive
c                      degrees clockwise from NS).
c   atol(ndir)       Azimuth (half window) tolerances
c   bandw(ndir)      Maximum bandwidth (i.e., the deviation in the
c                      direction perpendicular to the defined azimuth).
c   nvarg            Number of variograms to compute
c   ivtail(nvarg)    Variable for the tail of each variogram
c   ivhead(nvarg)    Variable for the head of each variogram
c   ivtype(nvarg)    Type of variogram to compute:
c                      1. semivariogram
c                      2. cross-semivariogram
c                      3. covariance
c                      4. correlogram
c                      5. general relative semivariogram
c                      6. pairwise relative semivariogram
c                      7. semivariogram of logarithms
c                      8. rodogram
c                      9. madogram
c                     10. indicator semivariogram: an indicator variable
c                         is constructed in the main program.
c
c
c
c OUTPUT VARIABLES:  The following arrays are ordered by direction,
c                    then variogram, and finally lag, i.e.,
c                      iloc = (id-1)*nvarg*(nlag+2)+(iv-1)*(nlag+2)+il
c
c   np()             Number of pairs
c   dis()            The average distance of pairs in this lag
c   gam()            Semivariogram, covariance, correlogram,... value
c   hm()             Mean of the tail data
c   tm()             Mean of the head data
c   hv()             Variance of the tail data
c   tv()             Variance of the head data
c
c
c
c PROGRAM NOTES:
c
c   1. The file "gamv2.inc" contains the dimensioning parameters.
c      These may have to be changed depending on the amount of data
c      and the requirements to compute different variograms.
c
c
c
c Original:  A.G. Journel                                           1978
c Revisions: K. Guertin                                             1980
c            C.V. Deutsch                                      July 1990
c-----------------------------------------------------------------------
      include  'gamv2.inc'
      parameter(PI=3.14159265)
      integer*4 nd,nlag,ndir,nvarg,nvar,nvario,i
      real*8    tmin,tmax,xlag,xltol
      real*8    x(nd),y(nd),u(nd,nvar),v(nd,nvar),
     +		azm(ndir),
     +          atol(ndir),bandw(ndir),uvxazm(MAXDIR),
     +          uvyazm(MAXDIR),csatol(MAXDIR)
      real*8    np(mxdlv),dis(mxdlv),gam(mxdlv),hm(mxdlv),
     +          tm(mxdlv),hv(mxdlv),tv(mxdlv)
      real*8    ivtail(nvarg),ivhead(nvarg),ivtype(nvarg)
      real*8    vr(maxdat,1)
      logical   omni 
      character option*8

c
c Define the distance tolerance if it isn't already:
c
      if(xltol.le.0.0) xltol = 0.5 * xlag
c
c Define the angles and tolerance for each direction:
c
      do 1 id=1,ndir
c
c The mathematical azimuth is measured counterclockwise from EW instead
c of clockwise from NS:
c
            azmuth     = (90.0-azm(id))*PI/180.0
            uvxazm(id) = cos(azmuth)
            uvyazm(id) = sin(azmuth)
            if(atol(id).le.0.0) then
                  csatol(id) = cos(45.0*PI/180.0)
            else
                  csatol(id) = cos(atol(id)*PI/180.0)
            endif
 1    continue
c
c Initialize the arrays for each direction, variogram, and lag:
c
      nsiz = ndir*nvarg*nlag+2
      if(nsiz.gt.MXDLV) then
            write(*,*) 'ERROR: available storage in gamv2 = ',MXDLV
            write(*,*) '       requested storage          = ',nsiz
            stop
      endif
      do 2 i=1,nsiz
            np(i)  = 0.
            dis(i) = 0.0
            gam(i) = 0.0
            hm(i)  = 0.0
            tm(i)  = 0.0
            hv(i)  = 0.0
            tv(i)  = 0.0
 2    continue
      dismxs = ((real(nlag) + 0.5 - EPSLON) * xlag) ** 2
c
c MAIN LOOP OVER ALL PAIRS:
c
      do 30 i=1,nd
      do 31 j=i,nd
c
c Definition of the lag corresponding to the current pair:
c
            dx = x(j) - x(i)
            dy = y(j) - y(i)
            hs = dx*dx+dy*dy
            if(hs.gt.dismxs) go to 31
            h  = sqrt(hs)
c
c Determine which lag this is and skip if outside the defined distance
c tolerance:
c
            if(h.le.EPSLON) then
                  il = 1
            else
                  il = int(h/xlag+0.5) + 2
                  if(il.gt.(nlag+2)) go to 31
                  if((abs(h-xlag*real(il-2))).gt.xltol) go to 31
            endif
c
c Definition of the direction corresponding to the current pair. All
c directions are considered (overlapping of direction tolerance cones
c is allowed):
c
            if(il.ne.1) then
                  uvxij = dx / h
                  uvyij = dy / h
            endif
            do 4 id=1,ndir
                  if(il.eq.1) then
                        dcazm = 0.0
                  else
                        dcazm = uvxij*uvxazm(id) + uvyij*uvyazm(id)
                        if((abs(dcazm)).lt.csatol(id)) go to 4
                  endif
c
c Check the bandwidth criteria (maximum deviation perpendicular to the
c specified direction azimuth):
c
                  band = uvxazm(id)*dy - uvyazm(id)*dx
                  if(abs(band).gt.bandw(id)) go to 4
c
c Check whether or not an omni-directional variogram is being computed:
c
                  omni = .false.
                  if(atol(id).ge.90.0) omni = .true.
c
c This direction is acceptable - go ahead and compute all variograms:
c
                  do 5 iv=1,nvarg
c
c For this variogram, sort out which is the tail and the head value:
c
                      it = ivtype(iv)
                      if(dcazm.ge.0.0) then
                            ii = ivtail(iv)
                            vrh   = vr(i,ii)
                            ii = ivhead(iv)
                            vrt   = vr(j,ii)
                            if(omni.or.it.eq.2) then
                                  ii    = ivhead(iv)
                                  vrtpr = vr(i,ii)
                                  ii    = ivtail(iv)
                                  vrhpr = vr(j,ii)
                            endif
                      else
                            ii = ivtail(iv)
                            vrh   = vr(j,ii)
                            ii = ivhead(iv)
                            vrt   = vr(i,ii)
                            if(omni.or.it.eq.2) then
                                  ii    = ivhead(iv)
                                  vrtpr = vr(j,ii)
                                  ii    = ivtail(iv)
                                  vrhpr = vr(i,ii)
                            endif
                      endif
c
c Reject this pair on the basis of missing values:
c
                      if(vrt.lt.tmin.or.vrh.lt.tmin.or.
     +                   vrt.ge.tmax.or.vrh.ge.tmax) go to 5
                      if(it.eq.2.and.(vrtpr.lt.tmin.or.vrhpr.lt.tmin.or.
     +                                vrtpr.ge.tmax.or.vrhpr.ge.tmax))
     +                                               go to 5
c
c We have an acceptable pair:
c
                      ii = (id-1)*nvarg*(nlag+2)+(iv-1)*(nlag+2)+il
c
c             COMPUTE THE APPROPRIATE "VARIOGRAM" MEASURE
c
c
c The Semivariogram:
c
      if(it.eq.1.or.it.eq.5.or.it.ge.10) then
            np(ii)  = np(ii)  + 1.
            dis(ii) = dis(ii) + dble(h)
            tm(ii)  = tm(ii)  + dble(vrt)
            hm(ii)  = hm(ii)  + dble(vrh)
            gam(ii) = gam(ii) + dble((vrh-vrt)*(vrh-vrt))
            if(omni) then
                  if(vrtpr.ge.tmin.and.vrhpr.ge.tmin.and.
     +               vrtpr.lt.tmax.and.vrhpr.lt.tmax) then
                        np(ii)  = np(ii)  + 1.
                        dis(ii) = dis(ii) + dble(h)
                        tm(ii)  = tm(ii)  + dble(vrtpr)
                        hm(ii)  = hm(ii)  + dble(vrhpr)
                        gam(ii) = gam(ii) + dble((vrhpr-vrtpr)*
     +                                           (vrhpr-vrtpr))
                  endif
            endif
c
c The Traditional Cross Semivariogram:
c
      else if(it.eq.2) then
            np(ii)  = np(ii)  + 1.
            dis(ii) = dis(ii) + dble(h)
            tm(ii)  = tm(ii)  + dble(0.5*(vrt+vrtpr))
            hm(ii)  = hm(ii)  + dble(0.5*(vrh+vrhpr))
            gam(ii) = gam(ii) + dble((vrhpr-vrh)*(vrt-vrtpr))
c
c The Covariance:
c
      else if(it.eq.3) then
            np(ii)  = np(ii)  + 1.
            dis(ii) = dis(ii) + dble(h)
            tm(ii)  = tm(ii)  + dble(vrt)
            hm(ii)  = hm(ii)  + dble(vrh)
            gam(ii) = gam(ii) + dble(vrh*vrt)
            if(omni) then
                  if(vrtpr.ge.tmin.and.vrhpr.ge.tmin.and.
     +               vrtpr.lt.tmax.and.vrhpr.lt.tmax) then
                        np(ii)  = np(ii)  + 1.
                        dis(ii) = dis(ii) + dble(h)
                        tm(ii)  = tm(ii)  + dble(vrtpr)
                        hm(ii)  = hm(ii)  + dble(vrhpr)
                        gam(ii) = gam(ii) + dble(vrhpr*vrtpr)
                  endif
            endif
c
c The Correlogram:
c
      else if(it.eq.4) then
            np(ii)  = np(ii)  + 1.
            dis(ii) = dis(ii) + dble(h)
            tm(ii)  = tm(ii)  + dble(vrt)
            hm(ii)  = hm(ii)  + dble(vrh)
            hv(ii)  = hv(ii)  + dble(vrh*vrh)
            tv(ii)  = tv(ii)  + dble(vrt*vrt)
            gam(ii) = gam(ii) + dble(vrh*vrt)
            if(omni) then
                  if(vrtpr.ge.tmin.and.vrhpr.ge.tmin.and.
     +               vrtpr.lt.tmax.and.vrhpr.lt.tmax) then
                        np(ii)  = np(ii)  + 1.
                        dis(ii) = dis(ii) + dble(h)
                        tm(ii)  = tm(ii)  + dble(vrtpr)
                        hm(ii)  = hm(ii)  + dble(vrhpr)
                        hv(ii)  = hv(ii)  + dble(vrhpr*vrhpr)
                        tv(ii)  = tv(ii)  + dble(vrtpr*vrtpr)
                        gam(ii) = gam(ii) + dble(vrhpr*vrtpr)
                  endif
            endif
c
c The Pairwise Relative:
c
      else if(it.eq.6) then
            if(abs(vrt+vrh).gt.EPSLON) then
                  np(ii)  = np(ii)  + 1.
                  dis(ii) = dis(ii) + dble(h)
                  tm(ii)  = tm(ii)  + dble(vrt)
                  hm(ii)  = hm(ii)  + dble(vrh)
                  gamma   = 2.0*(vrt-vrh)/(vrt+vrh)
                  gam(ii) = gam(ii) + dble(gamma*gamma)
            endif
            if(omni) then
                  if(vrtpr.ge.tmin.and.vrhpr.ge.tmin.and.
     +               vrtpr.lt.tmax.and.vrhpr.lt.tmax) then
                  if(abs(vrtpr+vrhpr).gt.EPSLON) then
                        np(ii)  = np(ii)  + 1.
                        dis(ii) = dis(ii) + dble(h)
                        tm(ii)  = tm(ii)  + dble(vrtpr)
                        hm(ii)  = hm(ii)  + dble(vrhpr)
                        gamma   = 2.0*(vrt-vrh)/(vrt+vrh)
                        gam(ii) = gam(ii) + dble(gamma*gamma)
                  endif
                  endif
            endif
c
c Variogram of Logarithms:
c
      else if(it.eq.7) then
            if(vrt.gt.EPSLON.and.vrh.gt.EPSLON) then
                  np(ii)  = np(ii)  + 1.
                  dis(ii) = dis(ii) + dble(h)
                  tm(ii)  = tm(ii)  + dble(vrt)
                  hm(ii)  = hm(ii)  + dble(vrh)
                  gamma   = alog(vrt)-alog(vrh)
                  gam(ii) = gam(ii) + dble(gamma*gamma)
            endif
            if(omni) then
                  if(vrtpr.ge.tmin.and.vrhpr.ge.tmin.and.
     +               vrtpr.lt.tmax.and.vrhpr.lt.tmax) then
                  if(vrtpr.gt.EPSLON.and.vrhpr.gt.EPSLON) then
                        np(ii)  = np(ii)  + 1.
                        dis(ii) = dis(ii) + dble(h)
                        tm(ii)  = tm(ii)  + dble(vrtpr)
                        hm(ii)  = hm(ii)  + dble(vrhpr)
                        gamma   = alog(vrt)-alog(vrh)
                        gam(ii) = gam(ii) + dble(gamma*gamma)
                  endif
                  endif
            endif
c
c Rodogram:
c
      else if(it.eq.8) then
            np(ii)  = np(ii)  + 1.
            dis(ii) = dis(ii) + dble(h)
            tm(ii)  = tm(ii)  + dble(vrt)
            hm(ii)  = hm(ii)  + dble(vrh)
            gam(ii) = gam(ii) + dble(sqrt(abs(vrh-vrt)))
            if(omni) then
                  if(vrtpr.ge.tmin.and.vrhpr.ge.tmin.and.
     +               vrtpr.lt.tmax.and.vrhpr.lt.tmax) then
                        np(ii)  = np(ii)  + 1.
                        dis(ii) = dis(ii) + dble(h)
                        tm(ii)  = tm(ii)  + dble(vrtpr)
                        hm(ii)  = hm(ii)  + dble(vrhpr)
                        gam(ii) = gam(ii) +dble(sqrt(abs(vrhpr-vrtpr)))
                  endif
            endif
c
c Madogram:
c
      else if(it.eq.9) then
            np(ii)  = np(ii)  + 1.
            dis(ii) = dis(ii) + dble(h)
            tm(ii)  = tm(ii)  + dble(vrt)
            hm(ii)  = hm(ii)  + dble(vrh)
            gam(ii) = gam(ii) + dble(abs(vrh-vrt))
            if(omni) then
                  if(vrtpr.ge.tmin.and.vrhpr.ge.tmin.and.
     +               vrtpr.lt.tmax.and.vrhpr.lt.tmax) then
                        np(ii)  = np(ii)  + 1.
                        dis(ii) = dis(ii) + dble(h)
                        tm(ii)  = tm(ii)  + dble(vrtpr)
                        hm(ii)  = hm(ii)  + dble(vrhpr)
                        gam(ii) = gam(ii) + dble(abs(vrhpr-vrtpr))
                  endif
            endif
      endif
c
c Finish loops over variograms, directions, and the double data loops:
c
 5              continue
 4          continue
 31   continue
 30   continue
c
c Get average values for gam, hm, tm, hv, and tv, then compute
c the correct "variogram" measure:
c
      do 6 id=1,ndir
      do 6 iv=1,nvarg
      do 6 il=1,nlag+2
            i = (id-1)*nvarg*(nlag+2)+(iv-1)*(nlag+2)+il
            if(np(i).le.0.) go to 6
            rnum   = np(i)
            dis(i) = dis(i) / dble(rnum)
            gam(i) = gam(i) / dble(rnum)
            hm(i)  = hm(i)  / dble(rnum)
            tm(i)  = tm(i)  / dble(rnum)
            hv(i)  = hv(i)  / dble(rnum)
            tv(i)  = tv(i)  / dble(rnum)
c
c  1. report the semivariogram rather than variogram
c  2. report the cross-semivariogram rather than variogram
c  3. the covariance requires "centering"
c  4. the correlogram requires centering and normalizing
c  5. general relative requires division by lag mean
c  6. report the semi(pairwise relative variogram)
c  7. report the semi(log variogram)
c  8. report the semi(rodogram)
c  9. report the semi(madogram)
c
            it = ivtype(iv)
            if(it.le.2) then
                  gam(i) = 0.5 * gam(i)
            else if(it.eq.3) then
                  gam(i) = gam(i) - hm(i)*tm(i)
            else if(it.eq.4) then
                  hv(i)  = dsqrt(hv(i)-hm(i)*hm(i))
                  tv(i)  = dsqrt(tv(i)-tm(i)*tm(i))
                  if((hv(i)*tv(i)).lt.EPSLON) then
                        gam(i) = 0.0
                  else
                        gam(i) =(gam(i)-hm(i)*tm(i))/(hv(i)*tv(i))
                  endif
c
c Square "hv" and "tv" so that we return the variance:
c
                  hv(i)  = hv(i)*hv(i)
                  tv(i)  = tv(i)*tv(i)
            else if(it.eq.5) then
                  htave  = real(0.5*(hm(i)+tm(i)))
                  htave  = htave   *   htave
                  if(htave.lt.EPSLON) then
                        gam(i) = 0.0
                  else
                        gam(i) = gam(i)/dble(htave)
                  endif
            else if(it.ge.6) then
                  gam(i) = 0.5 * gam(i)
            endif
 6    continue
      return
      end
